import * as THREE from 'https://unpkg.com/three@v0.150.0/build/three.module.js';

var renderer, scene, camera;
var chopper;

// We will load 4 textures into this array.
var textures = [];
var texturesToLoad = 4;
var textureIndex = 0; // Current texture index

var vertexShader, fragmentShader;

var lightPosition;
var lightTrajectory;

var viewerPosition = new THREE.Vector3(0.0, 2.0, 0.0); // Camera located 2 units in y direction.

function millis() {
    return (new Date()).getTime();
}

function toRad(degree) {
    return Math.PI * 2 * degree / 360;
}

window.onLoad = async () => {
    var canvasContainer = document.getElementById('myCanvasContainer');
    var width = 800;
    var height = 500;

    vertexShader = await (await fetch('vertex.glsl')).text();
    fragmentShader = await (await fetch('fragment.glsl')).text();

    renderer = new THREE.WebGLRenderer();
    renderer.setSize(width, height);
    canvasContainer.appendChild(renderer.domElement);

    scene = new THREE.Scene();

    // Setup the camera
    camera = new THREE.PerspectiveCamera(80, width / height, 1, 1000);
    camera.position.set(viewerPosition.x, viewerPosition.y, viewerPosition.z);
    camera.up = new THREE.Vector3(0, 0, -1); // The up-vector of the camera is the negative z
    camera.lookAt(new THREE.Vector3(0, 0, 0)); // We look at the origin
    scene.add(camera);

    lightTrajectory = new THREE.CatmullRomCurve3([
        new THREE.Vector3(-5, -7, 5),
        new THREE.Vector3(-4, -7, -2),
        new THREE.Vector3(3, -7, -4),
        new THREE.Vector3(8, -7, 6),
    ]);
    lightTrajectory.closed = true;

    window.addEventListener('keydown', function (event) {
        if (event.code == "ArrowLeft") {
            changeTexture();
        }
        if (event.code == "ArrowRight") {
            changeTexture();
        }
    }, false);

    // Add a small sphere, it will be located at the light position.
    // You probably did something like this already in the previous tasks for debugging.
    scene.add(new THREE.Mesh(new THREE.SphereGeometry(0.1, 6, 6), new THREE.MeshBasicMaterial({ color: 0xeeeeee })));

    // There is a loader to load the textures.
    var loader = new THREE.TextureLoader();
    loader.setCrossOrigin('Anonymous'); //This is needed because we are fetching files from another "server".
    // Local files do not work unless you have a web server running.
    loader.setPath('http://cglearn.eu/files/course/7/textures/');

    loader.load('sasuke256.png', onTextureLoaded);
    loader.load('ut256.png', onTextureLoaded);
    loader.load('wallTexture2.jpg', onTextureLoaded);
    loader.load('stripeTexture.png', onTextureLoaded);
}

/**
 * --Task-- This function is called when one texture is loaded.
 * Finish this function.
 */
function onTextureLoaded(texture) {
    // You may want to change some min and mag filters of the texture here.
    // You do want to make the texture repeating. Default is clamped at the edges.
    // After changing texture properties, tell it that it needs an update.

    // Push the texture to the array.

    if (--texturesToLoad == 0) {
        onAllTexturesLoaded();
    }
}

/**
 * All textures are loaded. Create and add the floor with the current texture. Start drawing the scene.
 */
function onAllTexturesLoaded() {
    addQuad(textures[textureIndex]);
    draw();
}

/**
 * This function changes the texture and updates it in the material.
 */
function changeTexture() {
    textureIndex = (textureIndex + 1) % 4
    scene.children[2].material.uniforms.colorTexture.value = textures[textureIndex];
    scene.children[2].material.needsUpdate = true;
    console.log('Current texture: ' + textureIndex);
}

function draw() {
    requestAnimationFrame(draw);

    var m = millis() / 4000;
    lightPosition = lightTrajectory.getPoint(m - parseInt(m));

    // Let's draw the sphere at the light's position.
    scene.children[1].position.set(lightPosition.x, lightPosition.y, lightPosition.z);

    scene.traverse(function (object) {
        if (object.material !== undefined && object.material instanceof THREE.ShaderMaterial) {
            object.material.uniforms.lightPosition.value = new THREE.Vector3().copy(lightPosition).applyMatrix4(camera.matrixWorldInverse);
        }
    });

    renderer.render(scene, camera);
}

/**
 * This assembles a quad with transforms and adds it to a scene.
 */
function addQuad(texture) {
    var halfPi = Math.PI / 2;

    var quad = createTexturedQuad(texture);
    quad.position.set(0, -10, 0);
    quad.rotation.set(-halfPi, 0, 0);

    scene.add(quad);
}

/**
 *  This function creates the actual quad mesh with the correct UV mapping.
 * 
 *  --Task--
 *  Finish this function.
 */
function createTexturedQuad(texture) {
    var geometry = new THREE.PlaneGeometry(20, 20, 1);

    // 1) Create a new Float32Array and set (array.set([...])) the UV coordinates in it

    // 2) Use the typed array to create a BufferAttribute
    //    and use geometry's setAttribute to assign it in the geometry.
    //    Name the geometry attribute as "uvCoord"

    var material = createShaderMaterial(texture);
    var wall = new THREE.Mesh(geometry, material);

    return wall;
}

/**
 *  This creates a shader material with the texture and lightPosition as uniforms.
 */
function createShaderMaterial(texture) {

    return new THREE.ShaderMaterial({
        uniforms: {
            lightPosition: {
                type: 'v3',
                value: lightPosition
            },
            colorTexture: {
                type: 't',
                value: texture
            }
        },
        vertexShader: vertexShader,
        fragmentShader: fragmentShader
    });
}
